@@ -40,12 +40,12 @@ typedef struct _GdaClientClass   GdaClientClass;
 typedef struct _GdaClientPrivate GdaClientPrivate;
 
 struct _GdaClient {
-	BonoboXObject object;
+	BonoboObject object;
 	GdaClientPrivate *priv;
 };
 
 struct _GdaClientClass {
-	BonoboXObjectClass parent_class;
+	BonoboObjectClass parent_class;
 	POA_GNOME_Database_Client__epv epv;
 
 	/* signals */
@@ -67,6 +67,7 @@ GdaConnection *gda_client_find_connection (GdaClient *client,
 					   const gchar *dsn,
 					   const gchar *username,
 					   const gchar *password);
+void           gda_client_close_all_connections (GdaClient *client);
 
 G_END_DECLS
 
