@@ -132,9 +132,10 @@ impl_Connection_close (PortableServer_Servant servant,
 		return TRUE;
 
 	result = gda_server_provider_close_connection (cnc->priv->provider, cnc);
-	if (!result)
+	if (!result) {
 		gda_error_list_to_exception (cnc->priv->errors, ev);
-	gda_server_connection_free_error_list (cnc);
+		gda_server_connection_free_error_list (cnc);
+	}
 
 	return result;
 }
@@ -142,7 +143,7 @@ impl_Connection_close (PortableServer_Servant servant,
 static GNOME_Database_RecordsetList *
 impl_Connection_executeCommand (PortableServer_Servant servant,
 				const GNOME_Database_Command *cmd,
-				const GNOME_Database_ParameterList *params,
+				const Bonobo_PropertySet *params,
 				CORBA_Environment *ev)
 {
 	GList *recset_list;
@@ -190,8 +191,10 @@ impl_Connection_beginTransaction (PortableServer_Servant servant,
 	bonobo_return_val_if_fail (GDA_IS_SERVER_CONNECTION (cnc), FALSE, ev);
 
 	result = gda_server_provider_begin_transaction (cnc->priv->provider, cnc, trans_id);
-	if (!result)
+	if (!result) {
 		gda_error_list_to_exception (cnc->priv->errors, ev);
+		gda_server_connection_free_error_list (cnc);
+	}
 	else {
 		gda_server_connection_notify_action (
 			cnc,
@@ -199,8 +202,6 @@ impl_Connection_beginTransaction (PortableServer_Servant servant,
 			NULL);
 	}
 
-	gda_server_connection_free_error_list (cnc);
-
 	return result;
 }
 
@@ -215,15 +216,16 @@ impl_Connection_commitTransaction (PortableServer_Servant servant,
 	bonobo_return_val_if_fail (GDA_IS_SERVER_CONNECTION (cnc), FALSE, ev);
 
 	result = gda_server_provider_commit_transaction (cnc->priv->provider, cnc, trans_id);
-	if (!result)
+	if (!result) {
 		gda_error_list_to_exception (cnc->priv->errors, ev);
+		gda_server_connection_free_error_list (cnc);
+	}
 	else {
 		gda_server_connection_notify_action (
 			cnc,
 			GNOME_Database_ACTION_TRANSACTION_FINISHED,
 			NULL);
 	}
-	gda_server_connection_free_error_list (cnc);
 
 	return result;
 }
@@ -239,15 +241,16 @@ impl_Connection_rollbackTransaction (PortableServer_Servant servant,
 	bonobo_return_val_if_fail (GDA_IS_SERVER_CONNECTION (cnc), FALSE, ev);
 
 	result = gda_server_provider_rollback_transaction (cnc->priv->provider, cnc, trans_id);
-	if (!result)
+	if (!result) {
 		gda_error_list_to_exception (cnc->priv->errors, ev);
+		gda_server_connection_free_error_list (cnc);
+	}
 	else {
 		gda_server_connection_notify_action (
 			cnc,
 			GNOME_Database_ACTION_TRANSACTION_ABORTED,
 			NULL);
 	}
-	gda_server_connection_free_error_list (cnc);
 
 	return result;
 }
@@ -269,7 +272,7 @@ impl_Connection_supports (PortableServer_Servant servant,
 static GNOME_Database_Recordset
 impl_Connection_getSchema (PortableServer_Servant servant,
 			   const GNOME_Database_Connection_Schema schema,
-			   const GNOME_Database_ParameterList *params,
+			   const Bonobo_PropertySet *params,
 			   CORBA_Environment *ev)
 {
 	GdaServerRecordset *recset;
@@ -278,13 +281,20 @@ impl_Connection_getSchema (PortableServer_Servant servant,
 
 	bonobo_return_val_if_fail (GDA_IS_SERVER_CONNECTION (cnc), CORBA_OBJECT_NIL, ev);
 
-	plist = gda_parameter_list_new_from_corba ((GNOME_Database_ParameterList *) params);
+	plist = gda_parameter_list_new_from_corba ((Bonobo_PropertySet *) params);
 	recset = gda_server_provider_get_schema (cnc->priv->provider, cnc, schema, plist);
 	gda_parameter_list_free (plist);
 
 	if (!GDA_IS_SERVER_RECORDSET (recset)) {
-		gda_error_list_to_exception (cnc->priv->errors, ev);
-		return CORBA_OBJECT_NIL;
+		if (cnc->priv->errors != NULL) {
+			gda_error_list_to_exception (cnc->priv->errors, ev);
+			gda_server_connection_free_error_list (cnc);
+
+			return CORBA_OBJECT_NIL;
+		}
+
+		/* create an empty recset, since there are no errors */
+		recset = gda_server_recordset_model_new (cnc, 0);
 	}
 
 	return bonobo_object_corba_objref (BONOBO_OBJECT (recset));
@@ -430,7 +440,7 @@ gda_server_connection_notify_action (GdaServerConnection *cnc,
 		client = (GNOME_Database_Client) l->data;
 		if (client != CORBA_OBJECT_NIL) {
 			CORBA_Environment ev;
-			GNOME_Database_ParameterList *corba_params;
+			Bonobo_PropertySet *corba_params;
 
 			corba_params = gda_parameter_list_to_corba (params);
 
@@ -469,15 +479,22 @@ gda_server_connection_add_error (GdaServerConnection *cnc, GdaError *error)
  * @gda_server_connection_add_error.
  */
 void
-gda_server_connection_add_error_string (GdaServerConnection *cnc, const gchar *msg)
+gda_server_connection_add_error_string (GdaServerConnection *cnc, const gchar *msg, ...)
 {
 	GdaError *error;
+	va_list args;
+        gchar sz[2048];
 
 	g_return_if_fail (GDA_IS_SERVER_CONNECTION (cnc));
 	g_return_if_fail (msg != NULL);
 
+	/* build the message string */
+	va_start (args, msg);
+        vsprintf (sz, msg, args);
+        va_end (args);
+
 	error = gda_error_new ();
-	gda_error_set_description (error, msg);
+	gda_error_set_description (error, sz);
 	gda_error_set_number (error, -1);
 	gda_error_set_source (error, g_get_prgname ());
 	gda_error_set_sqlstate (error, "-1");
