@@ -30,16 +30,18 @@ struct _GdaXmlDatabasePrivate {
 	gchar *uri;
 	GHashTable *tables;
 	GHashTable *views;
+	GHashTable *queries;
 };
 
-#define OBJECT_DATABASE    "database"
-#define OBJECT_DATA        "data"
-#define OBJECT_FIELD       "field"
-#define OBJECT_ROW         "row"
-#define OBJECT_TABLE       "table"
-#define OBJECT_TABLES_NODE "tables"
-#define OBJECT_VIEW        "view"
-#define OBJECT_VIEWS_NODE  "views"
+#define OBJECT_DATABASE     "database"
+#define OBJECT_DATA         "data"
+#define OBJECT_FIELD        "field"
+#define OBJECT_QUERIES_NODE "queries"
+#define OBJECT_ROW          "row"
+#define OBJECT_TABLE        "table"
+#define OBJECT_TABLES_NODE  "tables"
+#define OBJECT_VIEW         "view"
+#define OBJECT_VIEWS_NODE   "views"
 
 #define PROPERTY_ALLOW_NULL "isnull"
 #define PROPERTY_GDATYPE    "gdatype"
@@ -67,6 +69,11 @@ static GObjectClass *parent_class = NULL;
  * Private functions
  */
 
+static void
+process_queries_node (GdaXmlDatabase *xmldb, xmlNodePtr children)
+{
+}
+
 static void
 process_tables_node (GdaXmlDatabase *xmldb, xmlNodePtr children)
 {
@@ -118,6 +125,7 @@ gda_xml_database_init (GdaXmlDatabase *xmldb, GdaXmlDatabaseClass *klass)
 	xmldb->priv->uri = NULL;
 	xmldb->priv->tables = g_hash_table_new (g_str_hash, g_str_equal);
 	xmldb->priv->views = g_hash_table_new (g_str_hash, g_str_equal);
+	xmldb->priv->queries = g_hash_table_new (g_str_hash, g_str_equal);
 }
 
 static void
@@ -147,6 +155,9 @@ gda_xml_database_finalize (GObject *object)
 	g_hash_table_destroy (xmldb->priv->views);
 	xmldb->priv->views = NULL;
 
+	g_hash_table_destroy (xmldb->priv->queries);
+	xmldb->priv->queries = NULL;
+
 	g_free (xmldb->priv);
 	xmldb->priv = NULL;
 
@@ -225,6 +236,7 @@ gda_xml_database_new_from_uri (const gchar *uri)
 	}
 
 	xmldb = g_object_new (GDA_TYPE_XML_DATABASE, NULL);
+	xmldb->priv->uri = g_strdup (uri);
 
 	/* parse the file */
 	root = xmlDocGetRootElement (doc);
@@ -245,10 +257,12 @@ gda_xml_database_new_from_uri (const gchar *uri)
 			process_tables_node (xmldb, children);
 		else if (!strcmp (node->name, OBJECT_VIEWS_NODE))
 			process_views_node (xmldb, children);
+		else if (!strcmp (node->name, OBJECT_QUERIES_NODE))
+			process_queries_node (xmldb, children);
 		else {
-			gda_log_error (_("Invalid XML database file '%s'"), uri);
-			g_object_unref (G_OBJECT (xmldb));
-			return NULL;
+			//gda_log_error (_("Invalid XML database file '%s'"), uri);
+			//g_object_unref (G_OBJECT (xmldb));
+			//return NULL;
 		}
 
 		node = node->next;
@@ -257,6 +271,37 @@ gda_xml_database_new_from_uri (const gchar *uri)
 	return xmldb;
 }
 
+/**
+ * gda_xml_database_get_uri
+ * @xmldb: XML database.
+ *
+ * Return the URI associated with the given XML database. This URI will
+ * be used when saving the XML database (#gda_xml_database_save) and no
+ * URI is given.
+ *
+ * Returns: the URI for the XML database.
+ */
+const gchar *
+gda_xml_database_get_uri (GdaXmlDatabase *xmldb)
+{
+	g_return_val_if_fail (GDA_IS_XML_DATABASE (xmldb), NULL);
+	return xmldb->priv->uri;
+}
+
+/**
+ * gda_xml_database_set_uri
+ * @xmldb: XML database.
+ */
+void
+gda_xml_database_set_uri (GdaXmlDatabase *xmldb, const gchar *uri)
+{
+	g_return_if_fail (GDA_IS_XML_DATABASE (xmldb));
+
+	if (xmldb->priv->uri)
+		g_free (xmldb->priv->uri);
+	xmldb->priv->uri = g_strdup (uri);
+}
+
 /**
  * gda_xml_database_changed
  * @xmldb: XML database
@@ -272,6 +317,100 @@ gda_xml_database_changed (GdaXmlDatabase * xmldb)
 		       0);
 }
 
+/**
+ * gda_xml_database_reload
+ * @xmldb: XML database.
+ *
+ * Reload the given XML database from its original place, discarding
+ * all changes that may have happened.
+ */
+void
+gda_xml_database_reload (GdaXmlDatabase *xmldb)
+{
+	/* FIXME: implement */
+}
+
+/**
+ * gda_xml_database_save
+ * @xmldb: XML database.
+ * @uri: URI to save the XML database to.
+ *
+ * Save the given XML database to disk.
+ */
+gboolean
+gda_xml_database_save (GdaXmlDatabase *xmldb, const gchar *uri)
+{
+	g_return_val_if_fail (GDA_IS_XML_DATABASE (xmldb), FALSE);
+
+	/* FIXME: implement */
+	return FALSE;
+}
+
+static void
+add_table_to_list (gpointer key, gpointer value, gpointer user_data)
+{
+	gchar *name = (gchar *) key;
+	GList **list = (GList **) user_data;
+
+	*list = g_list_append (*list, g_strdup (name));
+}
+
+/**
+ * gda_xml_database_get_tables
+ * @xmldb: XML database.
+ *
+ * Return a list of all table names present in the given database.
+ * You must free the returned GList when you no longer need it, by
+ * using the #gda_xml_database_free_table_list function.
+ *
+ * Returns: a GList of strings.
+ */
+GList *
+gda_xml_database_get_tables (GdaXmlDatabase *xmldb)
+{
+	GList *list = NULL;
+
+	g_return_val_if_fail (GDA_IS_XML_DATABASE (xmldb), NULL);
+
+	g_hash_table_foreach (xmldb->priv->tables, (GHFunc) add_table_to_list, &list);
+	return list;
+}
+
+/**
+ * gda_xml_database_free_table_list
+ * @list: list of tables, as returned by #gda_xml_database_get_tables.
+ *
+ * Free a GList of strings returned by #gda_xml_database_get_tables.
+ */
+void
+gda_xml_database_free_table_list (GList *list)
+{
+	GList *l;
+
+	for (l = list; l != NULL; l = l->next)
+		g_free (l->data);
+	g_list_free (list);
+}
+
+/**
+ * gda_xml_database_find_table
+ * @xmldb: XML database.
+ * @name: Name for the table to look for.
+ *
+ * Searches the given XML database for a table named @name, and
+ * returns a pointer to it.
+ *
+ * Returns: a pointer to the table, or NULL if not found.
+ */
+GdaTable *
+gda_xml_database_find_table (GdaXmlDatabase *xmldb, const gchar *name)
+{
+	g_return_val_if_fail (GDA_IS_XML_DATABASE (xmldb), NULL);
+	g_return_val_if_fail (name != NULL, NULL);
+
+	return g_hash_table_lookup (xmldb->priv->tables, name);
+}
+
 /**
  * gda_xml_database_new_table
  * @xmldb: XML database.
@@ -328,8 +467,9 @@ gda_xml_database_new_table_from_node (GdaXmlDatabase *xmldb, xmlNodePtr node)
 		return NULL;
 	}
 
+	name = xmlGetProp (node, PROPERTY_NAME);
 	table = gda_table_new (name);
-	if (!table) {
+	if (table) {
 		gda_log_error (_("Table %s already exists"), name);
 		return NULL;
 	}
